@@ -1,7 +1,7 @@
 /*
  * Copyright 2000, International Business Machines Corporation and others.
  * All Rights Reserved.
- * 
+ *
  * This software has been released under the terms of the IBM Public
  * License.  For details, see the LICENSE file in the top-level source
  * directory or online at http://www.openafs.org/dl/license10.html
@@ -342,7 +342,7 @@ typedef union // WORKERPACKET
          PVOID hServer;	// [in] token from VosServerOpen()
          int idPartition;	// [in] partition ID
          LPTSTR pszVolume;	// [in] volume name
-         ULONG ckQuota;	// [in] initial quota
+         size_t ckQuota;	// [in] initial quota
          VOLUMEID idVolume;	// [out] volume ID
       } wpVosVolumeCreate;
 
@@ -454,7 +454,7 @@ typedef union // WORKERPACKET
          PVOID hServer;	// [in] token from VosServerOpen()
          int idPartition;	// [in] partition ID
          VOLUMEID idVolume;	// [in] id of read/write volume
-         ULONG ckQuota;	// [in] new quota
+         size_t ckQuota;	// [in] new quota
       } wpVosVolumeQuotaChange;
 
       struct {
@@ -470,7 +470,7 @@ typedef union // WORKERPACKET
       struct {
          PVOID hServer;	// [in] token from BosServerOpen()
          LPTSTR pszService;	// [in] name of new service
-         SERVICETYPE type;	// [in] service type (stCRON, stFS, etc)
+         AFSSERVICETYPE type;	// [in] service type (stCRON, stFS, etc)
          LPTSTR pszCommand;	// [in] full command-line to execute
          LPTSTR pszTimeCron;	// [in] date/time (CRON only)
          LPTSTR pszNotifier; 	// [in] command executed on exit
